/**
 * Copyright 2006 by jFramework.org.
 */
package org.jframework.core.util;

import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.lang.reflect.Method;

import org.apache.commons.lang.builder.EqualsBuilder;
import org.apache.commons.lang.builder.HashCodeBuilder;
import org.apache.commons.lang.builder.ToStringBuilder;

/**
 * A base class for all Java Beans.  The bean implements {@link #hashCode()}, {@link #equals(Object)}, and {@link #toString()} to examine bean properties through introspection.
 *
 * @author mgridnev
 * @version 2.0, Jul 28, 2006
 */
public class Bean
{
    /**
     * Returns hash code.
     *
     * @return hash code
     */
    public int hashCode()
    {
        final HashCodeBuilder builder = new HashCodeBuilder();
        enumerateProperties(new PropertyHandler()
        {
            public void handle(String name, Object value)
            {
                builder.append(value);
            }
        });

        return builder.toHashCode();
    }

    /**
     * Compares for equality.
     *
     * @param obj another object
     * @return <tt>true</tt> if obj is the same or its fields equal to this bean, <tt>false</tt> otherwise
     */
    public boolean equals(final Object obj)
    {
        if (obj == this) {
            return true;
        }

        if (obj == null || getClass() != obj.getClass()) {
            return false;
        }

        final EqualsBuilder builder = new EqualsBuilder();
        enumerateProperties(new PropertyHandler()
        {
            public void handle(String name, Object value)
            {
                Object theirs = BeanUtils.getProperty(obj, name);
                builder.append(value, theirs);

            }
        });

        return builder.isEquals();
    }

    /**
     * Returns string representation of this object.
     *
     * @return bean class name followed by a comma-separated list of propery name-value pairs
     */
    public String toString()
    {
        final ToStringBuilder builder = new ToStringBuilder(this);

        enumerateProperties(new PropertyHandler()
        {
            public void handle(String name, Object value)
            {
                builder.append(name, value);
            }
        });
        return builder.toString();
    }

    /**
     * Returns bean property descriptors.
     *
     * @return bean property descriptors
     */
    private PropertyDescriptor[] getPropertyDescriptors()
    {
        try {
            return Introspector.getBeanInfo(getClass()).getPropertyDescriptors();
        } catch (IntrospectionException e) {
            throw new InternalError("Introspection failed.");
        }
    }

    /**
     * Enumerates bean properties letting the given handler to act upon each property
     *
     * @param handler property handler
     */
    private void enumerateProperties(PropertyHandler handler)
    {
        for (PropertyDescriptor descriptor : getPropertyDescriptors()) {
            Method method = descriptor.getReadMethod();
            if (method != null && method.getDeclaringClass() != Object.class) {
                final String name = descriptor.getName();
                final Object value = BeanUtils.getProperty(this, name);
                handler.handle(name, value);
            }
        }
    }

    /**
     * Property handler.
     */
    private static interface PropertyHandler
    {
        /**
         * Handles property
         *
         * @param name  property name
         * @param value property value
         */
        void handle(String name, Object value);
    }
}
