/**
 * Copyright 2006-2007 by jFramework.org
 */
package org.jframework.core.util;

import java.beans.BeanInfo;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.lang.reflect.*;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.log4j.Logger;

/**
 * Unlike its more relaxed Apache commons <tt>BeanUtils</tt>, our implementation is a lot stricter. It perserves property type returned by {@link #getProperty(Object,String)} and
 * throws an exception if the property cannot be accessed.
 *
 * @author mgridnev
 * @version 2.0, March 1, 2007
 */
public abstract class BeanUtils
{
    /**
     * Logger.
     */
    private static final Logger log = Logger.getLogger(BeanUtils.class);

    /**
     * Setter prefix.
     */
    private static final String SETTER_PREFIX = "set";

    /**
     * Property name pattern.
     */
    private static final Pattern PROPERTY_NAME_PATTERN = Pattern.compile("([\\p{Alnum}_](\\[.*\\])?)+(\\.[\\p{Alnum}_]+(\\[.*\\])?)*");

    /**
     * A map of primitive types to wrapper types.
     */
    private static final Map<Class, Class> PRIMITIVE_TYPE_MAP;

    static {
        final Class[][] typeMap = {
            {byte.class, Byte.class},
            {short.class, Short.class},
            {int.class, Integer.class},
            {long.class, Long.class},
            {float.class, Float.class},
            {double.class, Double.class},
            {char.class, Character.class},
            {boolean.class, Boolean.class}
        };

        PRIMITIVE_TYPE_MAP = new HashMap<Class, Class>();

        for (Class[] types : typeMap) {
            PRIMITIVE_TYPE_MAP.put(types[0], types[1]);
        }
    }

    /**
     * Creates a new instance of the class.  The method is able to create an instance of private and protected classes.
     *
     * @param clazz bean class
     * @return a new instance
     */
    @SuppressWarnings("unchecked")
    public static <T> T newInstance(final Class<T> clazz, final Object... arguments)
    {
        final Constructor<T> ctor = (Constructor<T>) matchArguments(clazz.getDeclaredConstructors(), arguments);

        if (ctor == null) {
            throw new RuntimeException("Constructor matching " + Arrays.asList(arguments) + " cannot be found");
        }

        try {
            Accessible<Constructor> accessible = new Accessible<Constructor>(ctor);
            try {
                return ctor.newInstance(convertArguments(arguments, ctor.getParameterTypes()));
            } finally {
                accessible.restore();
            }
        } catch (Exception e) {
            throw new RuntimeException("Failed to create an instance of " + clazz.getName(), e);
        }
    }

    /**
     * Sets the bean property.
     *
     * @param bean     bean
     * @param fullName property name in dotted notation, e.g. <tt>"a.b.c"</tt>.  May also be an indexed property, e.g. <tt>"value[0]"</tt>.  In this case, the bean is expected to
     *                 have a setter <tt>void setValue(int, Object)</tt> or a getter <tt>List getValue()</tt>.   May also be a keyed property, e.g., <tt>"map[key]"</tt>.  In this
     *                 case, the bean is expected to have a getter <tt>Map getMap()</tt>.  The key is treated as a {@link String}.
     * @param values an array of property values. For a simple property including dotted notation, the bean is expected to have a setter <tt>setFoo(param, ..., paramN)</tt> where
     *                 the number of parameters matches the number of elements in the value array.  For indexed and keyed properties, the size of value array must be exactly one.
     * @throws RuntimeException if property is not defined or cannot be set
     */
    @SuppressWarnings("unchecked")
    public static void setProperty(final Object bean, final String fullName, final Object... values) throws RuntimeException
    {
        validatePropertyName(fullName);

        Object targetBean = getTargetBean(bean, fullName);
        PropertyName targetName = new PropertyName(getTargetProperty(fullName));

        try {
            Method method;
            Object[] arguments;

            if (targetName.getKey() == null) {
                // Simple property
                // NOTE: The following line is required because compiler gets confused between Object and Object[] args
                // So it calls this method instead of calling setProperty(Object, String, Object)
                if (values == null) {
                    arguments = new Object[]{null};
                } else {
                    arguments = values;
                }

                method = (Method) matchArguments(getSetters(targetBean.getClass(), targetName.getBaseName()), arguments);
            } else {
                // Keyed property as indexed accessor
                arguments = new Object[values.length + 1];
                System.arraycopy(values, 0, arguments, 1, values.length);
                arguments[0] = targetName.getKey();

                method = (Method) matchArguments(getSetters(targetBean.getClass(), targetName.getBaseName()), arguments);

                // Keyed property directly through the list
                if (method == null) {
                    arguments = null;
                    method = getReadMethod(targetBean, targetName.getBaseName(), null);
                }
            }

            if (method == null) {
                throw new RuntimeException("Property '" + bean.getClass().getName() + "#" + fullName + "' not found.");
            }

            Accessible<Method> accessible = new Accessible<Method>(method); // make sure we can access package private and private inner classes
            try {
                if (arguments != null) {
                    // Simple or indexed mutator
                    method.invoke(targetBean, convertArguments(arguments, method.getParameterTypes()));
                } else {
                    if (values.length != 1) {
                        throw new IllegalArgumentException("Invalid number of parameters.");
                    }

                    if (Map.class.isAssignableFrom(method.getReturnType())) {
                        // Directly through the map
                        final Map map = (Map) method.invoke(targetBean);
                        map.put(targetName.getKey(), values[0]);
                    } else {
                        // Directly through the list
                        final List list = (List) method.invoke(targetBean);
                        list.set(Integer.parseInt(targetName.getKey()), values[0]);
                    }
                }
            } finally {
                accessible.restore();
            }
        } catch (Exception e) {
            final String message = "Failed to set property '" + targetName.getBaseName() + "' = '" + Arrays.asList(values) + "' on " + targetBean.getClass().getName();
            log.error(message, e);
            throw new RuntimeException(message + ": cause = " + e.toString());
        }
    }

    /**
     * Parses a hierarchical property name to return the bean for the last property
     *
     * @param bean bean to start walking the property chain
     * @param name property name in dotted notation
     * @return a bean corresponding to the last property
     */
    private static Object getTargetBean(final Object bean, final String name)
    {
        final StringTokenizer tok = new StringTokenizer(name, ".");

        Object obj = bean;

        try {
            for (int cParts = tok.countTokens(); cParts > 1; cParts--) {
                final String part = tok.nextToken();

                final Object parent = obj;
                obj = getSingleProperty(parent, part);

                // If ancestor is null, try to initialize it
                if (obj == null) {
                    try {
                        obj = getPropertyType(parent, part).newInstance();
                        setProperty(parent, part, obj);
                    } catch (Exception e) {
                        log.error("Parent property is null and cannot be initialized", e);
                    }
                }
            }
        } catch (RuntimeException e) {
            log.error("Failed to get property " + name + " on " + bean.getClass().getName(), e);
            throw e;
        }

        return obj;
    }

    /**
     * Parses a hierarchical property name to return the last property.
     *
     * @param name property name in dotted notation
     * @return the last property name
     */
    private static String getTargetProperty(final String name)
    {
        final int sep = name.lastIndexOf('.');

        return sep < 0 ? name : name.substring(sep + 1);
    }

    /**
     * Validates property name
     *
     * @param name property name
     * @throws IllegalArgumentException if name is invalid
     */
    static void validatePropertyName(final String name)
    {
        if (!PROPERTY_NAME_PATTERN.matcher(name).matches()) {
            throw new IllegalArgumentException("Invalid property name: '" + name + "'");
        }
    }

    /**
     * Returns all setters corresponding to the given property.
     *
     * @param clazz    the class
     * @param property property name
     */
    private static Method[] getSetters(final Class clazz, final String property)
    {
        final Method[] methods = clazz.getMethods();
        final List<Method> setters = new LinkedList<Method>();

        for (Method method : methods) {
            if (isPropertySetter(method, property)) {
                setters.add(method);
            }
        }

        return setters.toArray(new Method[setters.size()]);
    }

    /**
     * Checks whether the specified method is a property setter.
     *
     * @param method   the method
     * @param property property name
     * @return <code>true</code> if the method is a setter, <code>false</code> otherwise
     */
    private static boolean isPropertySetter(final Method method, final String property)
    {
        final int iName = SETTER_PREFIX.length();

        final StringBuffer sbExpectedName = new StringBuffer(SETTER_PREFIX);
        sbExpectedName.append(property);
        sbExpectedName.setCharAt(iName, Character.toUpperCase(sbExpectedName.charAt(iName)));

        return sbExpectedName.toString().equals(method.getName());
    }

    /**
     * Searches for the best argument match in the list of methods or constructors.
     *
     * @param methodsOrCtors an array of {@link Method}s or {@link Constructor}s
     * @param arguments      arguments
     * @return the best match or <code>null</code> if no match can be found
     */
    private static Object matchArguments(final Object[] methodsOrCtors, final Object... arguments)
    {
        int bestScore = -1;
        Object methodBest = null;

        for (final Object method : methodsOrCtors) {
            final Class[] types;

            if (method instanceof Method) {
                types = ((Method) method).getParameterTypes();
            } else if (method instanceof Constructor) {
                types = ((Constructor) method).getParameterTypes();
            } else {
                throw new RuntimeException(method + " is neither method nor constructor");
            }

            final int score = matchArguments(types, arguments);

            if (score > bestScore) {
                methodBest = method;
                bestScore = score;
            } else if (score > -1 && score == bestScore) {
                throw new RuntimeException("Cannot resolve between " + methodBest + " and " + method);
            }
        }

        return methodBest;
    }

    /**
     * Checks if arguments match the parameter types.
     *
     * @param types  parameter types
     * @param values arguments
     * @return the score, the higher the score the closer the match. -1 means no match, 0 means not an exact match
     */
    private static int matchArguments(final Class[] types, final Object... values)
    {
        if (types.length != values.length) {
            return -1;
        }

        int score = 0;

        for (int i = 0; i < types.length; i++) {
            final Object value = values[i];

            if (value == null) {
                continue;
            }

            final Class<?> type = wrapPrimitiveType(types[i]);

            if (type.isAssignableFrom(value.getClass())) {
                score += 2; // direct match results in a higher score
                continue;
            }

            // Try as a ctor argument
            try {
                type.getConstructor(value.getClass());
                score++;
                continue;
            } catch (NoSuchMethodException e) {
            }

            // Try basic string conversion
            if (value instanceof String) {
                // To array
                if (type == char[].class || type == byte[].class) {
                    score++;
                    continue;
                }

                // To class name
                if (type == Class.class) {
                    try {
                        Class.forName((String) value);
                        score++;
                        continue;
                    } catch (Exception e) {
                    }
                }
            }

            return -1;
        }

        return score;
    }

    /**
     * Converts property arguments to parameter types.
     *
     * @param arguments      initial arguments
     * @param parameterTypes expected types
     * @return an array of correctly typed arguments
     * @throws RuntimeException if value conversion fails
     */
    static Object[] convertArguments(final Object[] arguments, final Class... parameterTypes)
        throws RuntimeException
    {
        if (arguments == null) {
            return new Object[0];
        }

        final Object[] typedArguments = new Object[arguments.length];

        for (int i = 0; i < typedArguments.length; i++) {
            final Class type = wrapPrimitiveType(parameterTypes[i]);
            final Object value = arguments[i];

            typedArguments[i] = convertArgument(value, type);
        }

        return typedArguments;
    }

    /**
     * Converts an argument to the given type.
     *
     * @param value source value
     * @param type  target type
     * @return converted value
     */
    public static Object convertArgument(final Object value, final Class<?> type)
    {
        if (value == null || type.isAssignableFrom(value.getClass())) {
            return value; // already of the correct type
        }

        // Attempt conversion

        // Direct
        try {
            final Constructor ctor = type.getConstructor(value.getClass());
            return ctor.newInstance(value);
        } catch (Exception e) {
        }

        // To string
        if (type == String.class) {
            try {
                final Method method = value.getClass().getMethod("toString");
                return method.invoke(value);
            } catch (Exception e) {
            }
        }

        // From string
        if (value instanceof String) {
            if (type == char[].class) {
                return ((String) value).toCharArray();
            }

            if (type == byte[].class) {
                return ((String) value).getBytes();
            }

            if (type == Class.class) {
                try {
                    return Class.forName((String) value);
                } catch (ClassNotFoundException e) {
                }
            }

            if (type == Date.class) {
                try {
                    return new Date(Long.parseLong((String) value));
                } catch (NumberFormatException e) {
                }
            }
        }

        final String message = "Value conversion failed for type " + type + " from " + value.getClass();
        log.error(message);
        throw new RuntimeException(message);
    }

    /**
     * Wraps a primitive types into its corresponding wrapper
     *
     * @param type primitive type
     * @return wrapper type
     */
    public static Class wrapPrimitiveType(final Class type)
    {
        final Class wrapperType = PRIMITIVE_TYPE_MAP.get(type);

        return wrapperType == null ? type : wrapperType;
    }

    /**
     * Return property type.
     *
     * @param bean     the bean
     * @param fullName property name in dotted notation
     * @return property class or <tt>null</tt> if the property is an indexed property
     */
    public static Class getPropertyType(Object bean, String fullName)
    {
        validatePropertyName(fullName);

        bean = getTargetBean(bean, fullName);
        fullName = getTargetProperty(fullName);

        PropertyName name = new PropertyName(fullName);

        return getReadMethod(bean, name.getBaseName(), name.getKey()).getReturnType();
    }

    /**
     * Returns bean property value preserving property type.
     *
     * @param bean     the bean
     * @param fullName property name in dotted notation, e.g. <tt>"a.b.c"</tt>.  May also be an indexed property, e.g. <tt>"value[0]"</tt>.  In this case, the bean is expected to
     *                 have a getter <tt>Object getValue(int)</tt> or a getter <tt>List getValue()</tt>.   May also be a keyed property, e.g., <tt>"map[key]"</tt>.  In this case,
     *                 the bean is expected to have a getter <tt>Map getMap()</tt>.  The key is treated as a {@link String}.
     * @return property value
     */
    public static Object getProperty(final Object bean, final String fullName)
    {
        validatePropertyName(fullName);

        Object obj = bean;

        try {
            for (StringTokenizer tok = new StringTokenizer(fullName, "."); tok.hasMoreTokens();) {
                final String part = tok.nextToken();

                if (obj == null) {
                    throw new NullPointerException("Encountered null ancestor of " + part + " when setting " + fullName + " on " +
                        bean.getClass().getName());
                }

                obj = getSingleProperty(obj, part);
            }
        } catch (RuntimeException e) {
            log.error("Failed to get property " + bean.getClass().getName() + "#" + fullName, e);
            throw e;
        }

        return obj;
    }


    /**
     * Returns property getter.
     *
     * @param bean the bean
     * @param name a single property name
     * @param key  property key, or <tt>null</tt> of this is not a keyed property
     * @return property read method
     */
    private static Method getReadMethod(final Object bean, final String name, String key)
    {
        try {
            final BeanInfo info = Introspector.getBeanInfo(bean.getClass());

            for (PropertyDescriptor pd : info.getPropertyDescriptors()) {
                if (!pd.getName().equals(name)) {
                    continue;
                }

                Method method = pd.getReadMethod();
                Class type = pd.getPropertyType();

                if (key != null) {
                    // Do we have an indexed accessor?
                    if (type == null) {
                        final char[] chars = name.toCharArray();
                        chars[0] = Character.toUpperCase(chars[0]);

                        try {
                            method = bean.getClass().getDeclaredMethod("get" + new String(chars), int.class);
                        } catch (NoSuchMethodException e) {
                            method = null;
                        }
                    } else if (!List.class.isAssignableFrom(type) && !Map.class.isAssignableFrom(type)) {
                        method = null;
                    }
                }

                // Did we find an appropriate accessor?
                if (method == null) {
                    throw new NullPointerException("No read method found for property.  Check property spelling.");
                }

                return method;
            }
            throw new RuntimeException("Property " + name + " not found");
        } catch (Exception e) {
            final String message = "Failed to find '" + name + "' getter on " + (bean == null ? null : bean.getClass().getName());
            log.error(message, e);
            throw new RuntimeException(message);
        }
    }

    /**
     * Returns bean property value preserving property type.
     *
     * @param bean     the bean
     * @param fullName a single property name
     * @return property value
     */
    private static Object getSingleProperty(final Object bean, final String fullName)
    {
        try {
            final PropertyName name = new PropertyName(fullName);
            final String key = name.getKey();

            final Method method = getReadMethod(bean, name.getBaseName(), key);
            Accessible<Method> accessible = new Accessible<Method>(method); // make sure we can access package private and  and private inner classes

            try {
                // Simple property
                if (key == null) {
                    return method.invoke(bean);
                }

                // Keyed property through list accessor
                if (List.class.isAssignableFrom(method.getReturnType())) {
                    final List list = (List) method.invoke(bean);
                    return list.get(Integer.parseInt(key));
                }

                // Keyed property though map accessor
                if (Map.class.isAssignableFrom(method.getReturnType())) {
                    final Map map = (Map) method.invoke(bean);
                    return map.get(key);
                }

                // Indexed property through indexed accessor
                return method.invoke(bean, Integer.parseInt(key));
            } finally {
                accessible.restore();
            }

        } catch (RuntimeException e) {
            throw e; // already processed
        } catch (Exception e) {
            final String message = "Failed to get property '" + fullName + "' on " + (bean == null ? null : bean.getClass().getName());
            log.error(message, e);
            throw new RuntimeException(message);
        }
    }

    /**
     * Property name
     */
    static class PropertyName
    {
        /**
         * Property name pattern
         */
        private static final Pattern PATTERN = Pattern.compile("([\\p{Alnum}_]+)(?:\\[(.*)\\])?");

        /**
         * Base name
         */
        final String baseName;

        /**
         * Property key
         */
        final String key;

        /**
         * Creates a property name
         */
        public PropertyName(String name)
        {
            Matcher matcher = PATTERN.matcher(name);

            if (!matcher.matches()) {
                throw new RuntimeException("Invalid property name: " + name);
            }

            baseName = matcher.group(1);
            this.key = matcher.group(2);
        }

        /**
         * Returns base property name
         */
        public String getBaseName()
        {
            return baseName;
        }

        /**
         * Returns property
         */
        public String getKey()
        {
            return key;
        }
    }

    /**
     * Controls accessiblity
     */
    static class Accessible<T extends AccessibleObject & Member>
    {
        /**
         * The accessible object
         */
        private final T object;

        /**
         * A flag indicating whether the object is publicly accessible
         */
        private final boolean isPublic;

        /**
         * A flag indicating whether the object is initially accessible
         */
        private final boolean wasAccessible;

        /**
         * Makes the object accessible
         */
        public Accessible(T object)
        {
            this.object = object;
            isPublic = Modifier.isPublic(object.getModifiers()) && Modifier.isPublic(object.getDeclaringClass().getModifiers());
            wasAccessible = object.isAccessible();

            if (!isPublic) {
                object.setAccessible(true);
            }
        }

        /**
         * Restore original accessible state
         */
        public void restore()
        {
            if (!isPublic) {
                object.setAccessible(wasAccessible);
            }
        }
    }
}
