/**
 * Copyright 2007 by jFramework.org.
 */
package org.jframework.core.util;

import net.sf.cglib.beans.BeanGenerator;
import net.sf.cglib.core.*;
import net.sf.cglib.proxy.*;
import net.sf.cglib.asm.ClassVisitor;
import net.sf.cglib.asm.Type;

import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.lang.reflect.Method;
import java.util.*;

import org.apache.log4j.Logger;

/**
 * Dynamic proxy utilities.
 *
 * @author mgridnev
 * @version 2.0, Jun 26, 2007
 */
public class DynamicProxyUtils
{
    /**
     * Logger
     */
    private static final Logger log = Logger.getLogger(DynamicProxyUtils.class);

    /**
     * Dynamically implements the bean specified by the interface.
     *
     * @param inf bean interface
     * @return bean instance
     */
    public static <T> T implement(final Class<? extends T> inf)
    {
        return implement(inf, null);
    }

    /**
     * Dynamically implements the bean specified by the interface.
     *
     * @param inf bean interface
     * @return bean instance
     */
    public static <T> T implement(final Class<? extends T> inf, final Class superclass)
    {
        if (!inf.isInterface()) {
            throw new IllegalArgumentException("Not an interface: " + inf.getName());
        }

        // Build complete property list
        final List<PropertyDescriptor> descriptors = new LinkedList<PropertyDescriptor>();

        try {
            enumerateProperties(descriptors, inf);
        } catch (IntrospectionException e) {
            log.error("Introspection failed");
            throw new RuntimeException("Failed to introspect " + inf.getName());
        }

        class Generator extends BeanGenerator
        {
            public Generator()
            {
                // Set the super class which is used by internal implementation to cache the generated class names
                setSuperclass(superclass);

                // Override the naming policy to add implemented interface to the class name
                setNamingPolicy(new DefaultNamingPolicy() {

                    public String getClassName(String prefix, String source, Object key, Predicate names)
                    {
                        prefix = (prefix == null ? "" : prefix + "$") + inf.getName();

                        return super.getClassName(prefix, source, key, names);
                    }
                });
            }

            public void generateClass(final ClassVisitor classVisitor) throws Exception
            {
                final ClassEmitter emitter = new ClassEmitter(classVisitor);
                emitter.begin_class(Constants.V1_2,
                    Constants.ACC_PUBLIC,
                    getClassName(),
                    superclass != null ? Type.getType(superclass) : Constants.TYPE_OBJECT,
                    new Type[]{Type.getType(inf)},
                    null);

                EmitUtils.null_constructor(emitter);

                // Generate property accessors
                for (final PropertyDescriptor descriptor : descriptors) {
                    generateAccessors(emitter, descriptor);
                }

                emitter.end_class();
            }
        }

        final BeanGenerator generator = new Generator();
        
        // IMPORTANT: Because of the way we overrode BeanGeneraator above we never actually add property names to the generator instance
        // The generator uses properties to calculate internal hash code which is used to cache already generated class names
        // If we leave caching on and don't add properties, we will start get a ClassCastException when attempting to cast to an implemented interface
        // Turning caching off will keep geneating a new class for every invocation which is not ideal either
        BeanGenerator.addProperties(generator, descriptors.toArray(new PropertyDescriptor[descriptors.size()]));

        @SuppressWarnings("unchecked") final
        T bean = (T) generator.create();

        // Initialize collection properties
        for (final PropertyDescriptor descriptor : descriptors) {
            final Class type = descriptor.getPropertyType();
            final Object value;

            if (type == List.class) {
                value = new LinkedList();
            } else if (type == Set.class) {
                value = new HashSet();
            } else if (type == Map.class) {
                value = new HashMap();
            } else if (type == Collection.class) {
                value = new LinkedList();
            } else {
                continue;
            }

            try {
                descriptor.getWriteMethod().invoke(bean, value);
            } catch (Exception e) {
                log.error("Failed to initialize collection property " + descriptor.getName(), e);
            }
        }

        return bean;
    }

    /**
     * Generates property accessors
     *
     * @param classEmitter class emitter
     * @param descriptor   property descriptor
     */
    private static void generateAccessors(final ClassEmitter classEmitter, final PropertyDescriptor descriptor)
    {
        // Field
        final String fieldName = "$cglib_prop_" + descriptor.getName();
        final Type type = Type.getType(descriptor.getPropertyType());
        classEmitter.declare_field(Constants.ACC_PRIVATE, fieldName, type, null, null);

        // Accessor
        CodeEmitter codeEmitter = classEmitter.begin_method(Constants.ACC_PUBLIC, new Signature(descriptor.getReadMethod().getName(), type, Constants.TYPES_EMPTY), null, null);
        codeEmitter.load_this();
        codeEmitter.getfield(fieldName);
        codeEmitter.return_value();
        codeEmitter.end_method();

        // Mutator
        codeEmitter = classEmitter.begin_method(Constants.ACC_PUBLIC, new Signature(descriptor.getWriteMethod().getName(), Type.VOID_TYPE, new Type[]{type}), null, null);
        codeEmitter.load_this();
        codeEmitter.load_arg(0);
        codeEmitter.putfield(fieldName);
        codeEmitter.return_value();
        codeEmitter.end_method();
    }


    /**
     * Recursively enumerates all interfaces in the class hierarchy
     *
     * @param descriptors a list of descriptor to populate with properties
     * @param cls         starting class
     * @throws IntrospectionException if an introspection error occurs
     */
    private static void enumerateProperties(final List<PropertyDescriptor> descriptors, final Class<?> cls) throws IntrospectionException
    {
        descriptors.addAll(Arrays.asList(Introspector.getBeanInfo(cls).getPropertyDescriptors()));

        for (final Class superinterface : cls.getInterfaces()) {
            enumerateProperties(descriptors, superinterface);
        }
    }

    /**
     * Decorates the original object using decorator.
     *
     * @param original  original object implementing at least one interface
     * @param decorator decorator object implementing at least one interface, may be <tt>null</tt>
     * @return decorated object implementing the combined interfaces of original and decorator objects
     */
    @SuppressWarnings("unchecked")
    public static <T> T decorate(final T original, final Object decorator)
    {
        final List<Class<?>> decoratorInterfaces = extractUnimplementedInterfaces(original, decorator);

        if (decoratorInterfaces.isEmpty()) {
            return original; // properly decorated already
        }

        final CallbackFilter filter = new CallbackFilter()
        {
            public int accept(final Method method)
            {
                return method.getDeclaringClass().isInstance(decorator) ? 0 : 1;
            }
        };

        final Callback decoratorCallback = new Dispatcher()
        {
            public Object loadObject()
            {
                return decorator;
            }
        };

        final Callback originalCallback = new Dispatcher()
        {
            public Object loadObject()
            {
                return original;
            }
        };

        final Class[] interfaces = decoratorInterfaces.toArray(new Class[decoratorInterfaces.size()]);

        return (T) Enhancer.create(original.getClass(), interfaces, filter, new Callback[]{decoratorCallback, originalCallback});
    }

    /**
     * Mixes two objects together.
     *
     * @param original   original object implementing at least one interface
     * @param decorator decorator object implementing at least one interface
     * @return decorated object implementing the combined interfaces of the two objects
     */
    public static Object mix(Object original, final Object decorator)
    {
        // First check if the original already implement all required decorator interfaces to avoid double decoration
        final List<Class<?>> decoratorInterfaces = extractUnimplementedInterfaces(original, decorator);

        if (decoratorInterfaces.isEmpty()) {
            return original; // properly decorated already
        }

        final List<Class<?>> originalInterfaces = Arrays.<Class<?>>asList(original.getClass().getInterfaces());

        // Collect all interfaces and make sure each interface corresponds to a delegate
        final List<Object> delegates = new LinkedList<Object>();

        for (int i = 0, size = originalInterfaces.size(); i < size; i++) {
            delegates.add(original);
        }

        for (int i = 0, size = decoratorInterfaces.size(); i < size; i++) {
            delegates.add(decorator);
        }

        final List<Class> interfaces = new LinkedList<Class>();
        interfaces.addAll(originalInterfaces);
        interfaces.addAll(decoratorInterfaces);

        // Mix them together
        return Mixin.create(interfaces.toArray(new Class[interfaces.size()]), delegates.toArray());
    }

    /**
     * Returns a list of interfaces not implemented by the original object
     *
     * @param original  an object implementing at least one interface
     * @param decorator an object implementing at least one interface
     * @return decorator interfaces not implemented by
     */
    static List<Class<?>> extractUnimplementedInterfaces(final Object original, final Object decorator)
    {
        // First check if the original already implement all required decorator interfaces to avoid double decoration
        final List<Class<?>> decoratorInterfaces = new ArrayList<Class<?>>(Arrays.<Class<?>>asList(decorator.getClass().getInterfaces()));

        // Remove already implemented interfaces
        for (ListIterator<Class<?>> it = decoratorInterfaces.listIterator(); it.hasNext();) {
            final Class<?> inf = it.next();
            if (inf.isInstance(original)) {
                it.remove();
            }
        }

        return decoratorInterfaces;
    }
}
