/**
 * Copyright 2008 by jFramework.org.
 */
package org.jframework.core.util;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Locale;

import org.apache.commons.lang.ObjectUtils;

/**
 * An implementation of {@link LocaleAware} that servers as a container for other locale aware objects. It propgates the changes to the locale to all of its children.
 *
 * @author mgridnev
 * @version 1.0, Jun 10, 2008
 */
public class LocaleAwareContainer implements LocaleAware
{
    /**
     * Locale.
     */
    private Locale locale = Locale.getDefault();

    /**
     * Contained children.
     */
    private final List<LocaleAware> children = new ArrayList<LocaleAware>();

    /**
     * {@inheritDoc}
     */
    public Locale getLocale()
    {
        return locale;
    }

    /**
     * {@inheritDoc}
     */
    public void setLocale(Locale locale)
    {
        if (ObjectUtils.equals(this.locale, locale)) {
            return; // already up to date
        }

        this.locale = locale;

        for (LocaleAware child : children) {
            child.setLocale(locale);
        }
    }

    /**
     * Returns container children.
     *
     * @return an immutable list
     */
    public List<LocaleAware> getChildren()
    {
        return Collections.unmodifiableList(children);
    }

    /**
     * Adds a child and synchronizes child's locale with the container's
     *
     * @param child locale aware child
     */
    public final void addChild(LocaleAware child)
    {
        if (child == null) {
            throw new NullPointerException("Child may not be null.");
        }

        child.setLocale(locale);

        children.add(child);
    }

    /**
     * Removes a child.
     *
     * @param child child to remove
     */
    public final void removeChild(LocaleAware child)
    {
        children.remove(child);
    }
}
