/**
 * Copyright 2006 by jFramework.org.
 */
package org.jframework.core.util;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.StringReader;
import java.math.BigInteger;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

/**
 * Number utilities.
 *
 * @author mgridnev
 * @version 1.0, Aug 11, 2006
 */
public abstract class NumberUtils
{
    /**
     * Parses a string which is a decimal unsigned long representation and returns its signed equivalent
     *
     * @param string string representation of a 64 bit unsigned long integer
     * @return 64 bit signed equivalent
     * @throws NumberFormatException if value is not a valid integer, a valid negative integer or an integer exceeding 64 bits
     */
    public static long parseUnsignedLong(final String string) throws NumberFormatException
    {
        return parseUnsignedLong(string, 10);
    }

    /**
     * Parses a string which is an unsigned long representation and returns its signed equivalent
     *
     * @param string string representation of a 64 bit unsigned long integer
     * @param radix  radix
     * @return 64 bit signed equivalent
     * @throws NumberFormatException if value is not a valid integer, a valid negative integer or an integer exceeding 64 bits
     */
    public static long parseUnsignedLong(final String string, final int radix) throws NumberFormatException
    {
        final BigInteger unsigned = new BigInteger(string, radix);

        if (unsigned.signum() < 0) {
            throw new NumberFormatException("Negative value.");
        }

        if (unsigned.bitLength() > 64) {
            throw new NumberFormatException("Exceeds 64 bits.");
        }

        return unsigned.longValue();
    }

    /**
     * Parses a string which is a decimal unsigned integer representation and returns its signed equivalent
     *
     * @param string string representation of a 64 bit unsigned integer integer
     * @return 64 bit signed equivalent
     * @throws NumberFormatException if value is not a valid integer, a valid negative integer or an integer exceeding 64 bits
     */
    public static int parseUnsignedInt(final String string) throws NumberFormatException
    {
        return parseUnsignedInt(string, 10);
    }

    /**
     * Parses a string which is an unsigned integer representation and returns its signed equivalent
     *
     * @param string string representation of a 64 bit unsigned integer integer
     * @param radix  radix
     * @return 64 bit signed equivalent
     * @throws NumberFormatException if value is not a valid integer, a valid negative integer or an integer exceeding 64 bits
     */
    public static int parseUnsignedInt(final String string, final int radix) throws NumberFormatException
    {
        final BigInteger unsigned = new BigInteger(string, radix);

        if (unsigned.signum() < 0) {
            throw new NumberFormatException("Negative value");
        }

        if (unsigned.bitLength() > 32) {
            throw new NumberFormatException("Exceeds 32 bits");
        }

        return unsigned.intValue();
    }


    /**
     * Returns a deciaml string representation of a 64 bit unsigned long integer
     *
     * @param signed 64 bit singed equivalent
     * @return string representation
     */
    public static String toUnsignedLongString(final long signed)
    {
        return toUnsignedLongString(signed, 10);
    }

    /**
     * Returns a string representation of a 64 bit unsigned long integer
     *
     * @param signed 64 bit singed equivalent
     * @param radix  radix
     * @return string representation
     */
    public static String toUnsignedLongString(final long signed, final int radix)
    {
        // IMPORTANT: Using hex so sign is ignored
        return new BigInteger(Long.toHexString(signed), 16).toString(radix);
    }

    /**
     * Returns a deciaml string representation of a 64 bit unsigned integer integer
     *
     * @param signed 64 bit singed equivalent
     * @return string representation
     */
    public static String toUnsignedIntString(final int signed)
    {
        return toUnsignedIntString(signed, 10);
    }

    /**
     * Returns a string representation of a 64 bit unsigned integer integer
     *
     * @param signed 64 bit singed equivalent
     * @param radix  radix
     * @return string representation
     */
    public static String toUnsignedIntString(final int signed, final int radix)
    {
        // IMPORTANT: Using hex so sign is ignored
        return new BigInteger(Integer.toHexString(signed), 16).toString(radix);
    }

    /**
     * Returns a hexadecimal representation of unsigned byte array
     *
     * @param rgBytes byte array
     * @return string representation
     */
    public static String toByteArrayString(final byte[] rgBytes)
    {
        final StringBuffer sb = new StringBuffer();
        for (byte value : rgBytes) {
            final String string = Integer.toHexString(value & 0x00ff);

            if (string.length() == 1) {
                sb.append('0');
            }

            sb.append(string);
        }

        return sb.toString();
    }

    /**
     * Parses a hexadecimal representation of unsigned byte array
     *
     * @param string encoded byte array
     * @return byte array
     * @throws NumberFormatException if string does not represent a valid byte array
     */
    public static byte[] parseByteArrayString(final String string) throws NumberFormatException
    {
        final ByteArrayOutputStream out = new ByteArrayOutputStream();

        final StringReader reader = new StringReader(string);
        final char[] chByte = new char[2];

        try {
            while (true) {
                final int read = reader.read(chByte);

                if (read == -1) {
                    return out.toByteArray(); // EOF
                }

                if (read == chByte.length) {
                    out.write(Integer.parseInt(new String(chByte), 16));
                    continue;
                }

                throw new NumberFormatException("Not a byte array: " + string);
            }
        } catch (IOException e) {
            throw new InternalError(); // should never happen
        }
    }

    /**
     * Calculates MD5 hash over given byte array
     */
    public static byte[] MD5(final byte[] rgInput)
    {
        final MessageDigest digest;

        try {
            digest = MessageDigest.getInstance("MD5");
        } catch (NoSuchAlgorithmException e) {
            throw new InternalError("Failed to get MD5 digester"); // should never happen
        }

        digest.update(rgInput);

        return digest.digest();
    }
}
